/*****************************************************
 *                    VN.C
 *
 *        A .IMQ / .IMG NASA Files Viewer
 *
 *            version Linux-950114
 *
 * This program is free software and is distributed
 * under the terms of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This version requires a SVGA video card with the
 * 1024x768x256 video mode and the SVGALIB library.
 *****************************************************/

#include <ctype.h>
#include <dirent.h>
#include <getopt.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
#include <math.h>
#include <errno.h>
#include "vn.h"

#ifndef INSTALLDIR
#define INSTALLDIR "/usr/local/lib/vn"
#endif

#define FONTPATH INSTALLDIR"/fonts"

typedef struct {
	int key_scancode;
	int func_num;
}

CMD;

char *GNUcopyleft =
"VN is free software and is distributed under the terms of the GNU\n"
"General Public License as published by the Free Software Foundation.\n"
"VN comes with ABSOLUTELY NO WARRANTY; see the file COPYING for details.\n\n";


CMD Commands[] =
{
	{0x3b, 1},		/* F1 */
	{0x3c, 2},		/* F2 */
	{0x3d, 3},		/* F3 */
	{0x3e, 4},		/* F4 */
	{0x3f, 5},		/* F5 */
	{0x40, 6},		/* F6 */
	{0x41, 7},		/* F7 */
	{0x42, 8},		/* F8 */
	{0x43, 9},		/* F9 */
	{0x44, 10},		/* F10 */
	{0, 0}
};

/*
 * Buffers for character fonts
 */
char *small_font;
char *normal_font;
char *big_font;

int have_normalfont = 0, have_bigfont = 0;

/*
 * These functions are in DECOMP.C
 */
extern void free_tree(long int *nfreed);
extern void decmpinit(long int *hist);
extern void decompress(char *ibuf, char *obuf, long int *nin, long int *nout);

/*
 * This function is in DISPENG.C
 */
extern void disp_engtable(engtable * engtab, int part, int disable_clipping);

/*
 * This function is in DISPSUFX.C
 */
extern void dispsufx(line_suffix linsufx[], int part, int disable_clipping);

/*
 * Version string is defined in VERSION.C
 */
extern char version[];

/*
 * Global variables
 */
char *image_info[200];
char putbuf[1000 * 27 + 4], tbuf[2048], linei[1000], tline[132];
char filter[30], target[30], spacecraft[30], date[30], camera[30];
engtable engtab;
line_suffix linsufx[800];
unsigned char palette[2][256][3];
long image_hist[256], encoding_hist[512];
int what_left;
extern int optind, opterr, optopt;
extern char *optarg;
struct option lopts[] =
{
	{"auto_display", no_argument, 0, 'a'},
	{"auto_display_delay", required_argument, 0, 'd'},
	{"batch", no_argument, 0, 'b'},
	{"clearscreen", no_argument, 0, 'c'},
	{"default_info", required_argument, 0, 'i'},
	{"default_palette", required_argument, 0, 'p'},
	{"disable_image_scaling", no_argument, 0, 's'},
	{"disable_text_clipping", no_argument, 0, 't'},
	{"help", no_argument, 0, 'h'},
	{NULL, 0, 0, 0}
};

int get_record(FILE * fp, char *buffer, int length)
{
	int len;

	if (length == 0) {
		len = 0;
		fread(&len, 2, 1, fp);
		fread(buffer, 1, len + len % 2, fp);
	} else {
		len = length;
		fread(buffer, 1, len, fp);
	}
	return len;
}


#if defined(NO_BLANKS) | defined(LESS_BLANKS)

#ifdef NO_BLANKS
#define BLANKS_TO_RIP 80
#else
#define BLANKS_TO_RIP 20
#endif

/*
 * replace terminating end of line marker(s) with null and skip leading
 * and trailing spaces
 */
char *rip(char *s)
{
	char *cp;

	if ((cp = strchr(s, '\r')) != NULL)
		*cp = '\0';
	if ((cp = strchr(s, '\n')) != NULL)
		*cp = '\0';

	for (cp = s + strlen(s) - 1; *cp == ' '; cp--);
	*(cp + 1) = '\0';

	for (cp = s; *cp == ' '; cp++);

	return cp;
}

/*
 * trim extra spaces
 */
char *rip_spaces(char *s)
{
	char *ts1, *ts2, *cp;
	int n, done;

	ts2 = s;
	done = 0;
	n = 0;
	ts1 = strdup(s);

	for (cp = ts1; *cp != '\0'; cp++) {
		*ts2++ = *cp;
		if (*cp == ' ' && !done) {
			while ((*cp == ' ') && (n < BLANKS_TO_RIP)) {
				cp++;
				n++;
			}
			*ts2++ = *cp;
			done = ((n >= BLANKS_TO_RIP) || (*cp == '\0'));
		}
	}
	*ts2 = '\0';
	free(ts1);

	return s;
}

#endif

/*
 * graphic screen input routine
 */
char *graphgets(char *prompt, char *s, int len, int numonly)
{
	char ch, *cp, sch[2];
	char buf[80];
	int n = 0;

	gl_setclippingwindow(0, 700, 223, 740);
	gl_clearscreen(0);
	if (have_normalfont)
		gl_setfont(8, 16, normal_font);
	gl_write(0, 720, prompt);

	keyboard_close();

	sch[1] = '\0';
	bzero(s, strlen(s));
	cp = s;
	while ((ch = vga_getch()) != '\n' && ((n < len) || (ch == '\b'))) {
		if (ch == '\b' || ch == '\x7f') {
			if (n > 0) {
				n--;
				*--cp = '\0';
				gl_clearscreen(0);
				sprintf(buf, "%s%s", prompt, s);
				gl_write(0, 720, buf);
			}
			continue;
		}
		if (!numonly || isdigit(ch) || ch == '.') {
			n++;
			*cp++ = ch;
			sprintf(buf, "%s%s", prompt, s);
			gl_write(0, 720, buf);
		}
	}
	*cp = '\0';
	keyboard_init();
	gl_setfont(8, 8, small_font);
	gl_clearscreen(0);
	return s;
}


void disp_info(int disable_clipping)
{

	gl_setclippingwindow(0, 75, 223, 740);
	gl_clearscreen(0);
	if (disable_clipping)
		gl_setclippingwindow(0, 75, 1023, 250);
	if (have_normalfont)
		gl_setfont(8, 16, normal_font);
	gl_write(0, 105, date);
	gl_write(0, 130, spacecraft);
	gl_write(0, 155, target);
	gl_write(0, 180, camera);
	gl_write(0, 205, filter);
	gl_setfont(8, 8, small_font);

}


void display_data(int start, int end, int disable_clipping)
{
	int il;
	char buf[128];

	gl_setclippingwindow(0, 75, 223, 740);
	gl_clearscreen(0);
	if (disable_clipping)
		gl_setclippingwindow(0, 75, 1023, 740);
	gl_setfont(8, 8, small_font);
	for (il = start; il < end; il++) {
#if defined(NO_BLANKS) | defined(LESS_BLANKS)
		strcpy(buf, image_info[il]);
		rip_spaces(buf);
		gl_write(0, 12 * il + 80, buf);
#else
		gl_write(0, 12 * il + 80, image_info[il]);
#endif
	}
	gl_setclippingwindow(0, 0, 767, 1023);
}

/*
 * display image histogram
 */
void display_hist(void)
{
	int n, x;
	long total, max;
	int nmax, nmin, nmax1, nmin1;
	char buf[80];

	gl_setclippingwindow(0, 75, 223, 740);
	gl_clearscreen(0);
	gl_setfont(8, 8, small_font);
	nmax = 0;
	nmin = 0;
	total = 0l;
	max = 0l;
	for (n = 0; n < 256; n++) {
		total += image_hist[n];
		if (image_hist[n] > max) {
			max = image_hist[n];
			nmax = n;
		}
	}
	for (n = 0; n < 256; n++) {
		if (((n & 0x0f) == 0) || (n == 255)) {
			sprintf(buf, "%3d", n);
			gl_write(0, n * 2 + 96, buf);
			gl_hline(22, n * 2 + 100, 25, 255);
		}
		x = 200l * image_hist[n] / max + 25;
		gl_hline(25, n * 2 + 100, x, 255);
	}
	nmax1 = nmin1 = 0;
	for (n = 0; n < 256; n++) {
		if (image_hist[n] > (max / 100)) {
			nmax1 = n;
		}
	}
	for (n = 255; n >= 0; n--) {
		if (image_hist[n] > (max / 100)) {
			nmin1 = n;
		}
	}
	if (have_normalfont)
		gl_setfont(8, 16, normal_font);
	sprintf(buf, "Total of %lu pixels", total);
	gl_write(0, 650, buf);
	sprintf(buf, "Max at %d = %lu (%.2f%%) ", nmax, max, (100.0 * max / total));
	gl_write(0, 666, buf);
	sprintf(buf, "1%% points: n1 = %d, n2 = %d", nmin1, nmax1);
	gl_write(0, 682, buf);
}

/*
 * display help
 */
void help(void)
{
	int y = 100;

	if (have_normalfont)
		gl_setfont(8, 16, normal_font);
	gl_setclippingwindow(0, y, 223, 740);
	gl_clearscreen(0);
	gl_write(0, y, "**** Available Commands ****");
	gl_write(0, y += 50, "      F1  Help");
	gl_write(0, y += 20, "Shift+F1  Standard info");
	gl_write(0, y += 20, "      F2  Image info");
	gl_write(0, y += 20, "Shift+F2 *Image histogram");
	gl_write(0, y += 20, "      F3 *Engin. table info");
	gl_write(0, y += 20, "      F4 *Next line info");
	gl_write(0, y += 20, "Shift+F4 *Selected line info");
	gl_write(0, y += 20, "      F5  Set palette  #1");
	gl_write(0, y += 20, "Shift+F5  Load palette #1");
	gl_write(0, y += 20, "      F6  Set palette  #2");
	gl_write(0, y += 20, "Shift+F6  Load palette #2");
	gl_write(0, y += 20, "      F7  Enhance contrast");
	gl_write(0, y += 20, "Shift+F7  Exp/Log enhance");
	gl_write(0, y += 20, "      F8  Reload image");
	gl_write(0, y += 20, "Shift+F8  Toggle scaling");
	gl_write(0, y += 20, "      F9  Next image");
	gl_write(0, y += 20, "Shift+F9  Autoshow toggle");
	gl_write(0, y += 20, "      F10 Quit program");
	gl_write(0, y += 40, "    * = IMQ images only");
	gl_write(0, y += 40, "VN is  free software  and is");
	gl_write(0, y += 20, "distributed under  the terms");
	gl_write(0, y += 20, "of the GNU GPL, published by");
	gl_write(0, y += 20, "the Free Software Foundation");
	gl_write(0, y += 40, "VN comes with  ABSOLUTELY NO");
	gl_write(0, y += 20, "WARRANTY; see the file");
	gl_write(0, y += 20, "COPYING for details.");
	gl_setfont(8, 8, small_font);
}

/*
 * display usage info
 */
void usage(void)
{
	printf("\nNASA IMQ/IMG FILES VIEVER\n");
	printf("%s\n\n", version);
	printf(GNUcopyleft);
	printf("Usage: vn [<options>] <file>\n\n"
	       "Valid options are:\n"
	       "\t--auto_display, -a\n"
	       "\t--auto_delay <seconds>, -d <seconds>\n"
	       "\t--batch, -b\n"
	       "\t--clearscreen, -c\n"
	       "\t--default_info <info #>, -i <info #>\n"
	       "\t--default_palette <filename>, -p <filename>\n"
	       "\t--disable_image_scaling, -s\n"
	       "\t--disable_text_clipping, -t\n"
	       "\t--help, -h\n\n");
	printf("<file> may contain wildcard characters  (i.e. /SUBDIR/C2063*.*).\n\n");
	printf("N.B. To run, this program needs a SVGA card supporting the 1024x768x256 mode\n\n");
}

/*
 * read a palette file into a buffer
 */
int read_palette(char *filename, unsigned char *palette)
{
	char fname[80], s[82];
	FILE *pf;
	int n, r, g, b;

	/* read a palette file into palette array */

	sprintf(fname, "%s/palettes/%s", INSTALLDIR, filename);

#ifdef DEBUG
	printf("Attempting to open %s\n", fname);
#endif
	if ((pf = fopen(fname, "r")) == NULL) {
		return -1;
	}
	/* skip label */

	do {
		fgets(s, 80, pf);
		if (strstr(s, "TABLE_ROWS") != NULL) {
			n = atoi(strchr(s, '=') + 2);
			if (n != 256) {
				gl_write(0, 750, "Not a 256-color map");
				return -1;
			}
		}
	}
	while (strncmp(s, "END", 3));

	for (n = 0; n != 255;) {
		fgets(s, 80, pf);
		sscanf(s, "%d %d %d %d\n", &n, &r, &g, &b);
		*(palette + 3 * n) = r;
		*(palette + 3 * n + 1) = g;
		*(palette + 3 * n + 2) = b;
#ifdef DEBUG
		printf("color #%d: r=%d g=%d b=%d\n", n, r, g, b);
#endif
	}
	fclose(pf);
	return 0;
}

/*
 * display no more files prompt
 */
void no_more_files(int more, int batch_mode)
{
	gl_setclippingwindow(0, 75, 223, 740);
	gl_clearscreen(0);
	gl_setfont(8, 16, normal_font);
	if (more)
		gl_write(50, 300, "No more images!");
	else
		gl_write(75, 300, "No images!");
	gl_setfont(8, 8, small_font);
	keyboard_close();
	if (!batch_mode) {
		gl_write(25, 320, "Press any key to exit");
		getchar();
	}
}

/*
 * redisplay left side info
 */
void restore_left(int what_left, int nline, int line, int part, int disable_clipping)
{

	switch (what_left) {
	case 0:
		disp_info(disable_clipping);
		break;
	case 1:
		help();
		break;
	case 2:
		display_data(0, nline, disable_clipping);
		break;
	case 3:
		disp_engtable(&engtab, part, disable_clipping);
		break;
	case 4:
		dispsufx(linsufx, line, disable_clipping);
		break;
	case 5:
		display_hist();
		break;

	}
}

/*
 * convert a n-pixel-by-n-bits buffer into a n-bytes one
 */
void convert(char *ibuf, char *obuf, int nbytes, int nbits)
{
	int n, m, numnibbles;
	char mask, byte, nibble;
	char *ip, *op;

	if (nbits == 8) {
		memcpy(obuf, ibuf, nbytes);
	} else {
		numnibbles = 8 / nbits;
		mask = (-128 >> (nbits - 1));
		ip = ibuf;
		op = obuf;
		for (n = 0; n < nbytes; n++, ip++) {
			byte = *ip;
			for (m = 0; m < numnibbles; m++) {
				nibble = ((byte & mask) >> (8 - nbits));
				byte = byte << nbits;
				*op++ = ((1 << nbits) - 1) * nibble;
			}
		}
	}
}

/*
 * get a command key
 */
int get_command(void)
{

	int shift;
	CMD *cmdp;

	keyboard_update();
	for (cmdp = Commands; cmdp->key_scancode != 0; cmdp++) {
		if (keyboard_keypressed(cmdp->key_scancode)) {
			shift = (keyboard_keypressed(42) != 0) || keyboard_keypressed(54) != 0;
#ifdef DEBUG
			printf("*** Scancode = %d, Cmd = %d Shift = %d\n", cmdp->key_scancode, cmdp->func_num, shift);
#endif
			return (cmdp->func_num) + ((shift) ? 10 : 0);
		}
	}
	return -1;
}

/*
 * restore signals to its original state
 */
void restoresignal(void)
{
	signal(SIGHUP, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
}

/*
 * terminate program
 */
void vn_close(int n)
{
	keyboard_close();
	vga_setmode(TEXT);
	restoresignal();
	exit(0);
}

/*
 * set signal handlers
 */
void setsignal(void)
{
	signal(SIGHUP, vn_close);
	signal(SIGTERM, vn_close);
}

/*
 * main program
 */
int main(int argc, char *argv[])
{

	FILE *fp;
	char *cp, temp[80], fname[80], lblname[80], st[80], def_pal[80];
	char *tmpbuf[4096];
	clock_t now, then;
	int autodisp, len, nline, npix, end, end_all, part, line, bytestoskip;
	int ch, disable_scaling = 0, disable_clipping = 0, clearscreen = 0;
	int batch_mode = 0, enhanced;
	int image_start, hist_start, detached_label, imgfile, ibgfile,
	  loptind;
	int c, n, nmin, nmax, x, y, ymax, delay;
	int newpixel, oldpixel;
	int record_bytes, frecords, lrecords, lbytes, imqfile, nbits;
	long max, min;
	long nsi, nso, image_lines, il;
	long num;
	float k, s, r;


	setsignal();

	autodisp = 0;
	what_left = 0;
	part = 0;
	line = -2;
	def_pal[0] = '\0';
	delay = 1l;

	for (y = 1; y < 200; y++)
		image_info[y] = "Not available";

	/*
	 * process options
	 */
	while ((c = getopt_long(argc, argv, "abcd:hi:p:st", lopts, &loptind)) != -1) {
		switch (c) {
		case 'a':
			autodisp = ~0;
			break;
		case 'b':
			batch_mode = ~0;
			break;
		case 'c':
			clearscreen = ~0;
			break;
		case 'd':
			delay = atoi(optarg);
			break;
		case 'h':
			usage();
			exit(0);
		case 'i':
			what_left = atoi(optarg) % 6;
			break;
		case 's':
			disable_scaling = ~0;
			break;
		case 't':
			disable_clipping = ~0;
			break;
		case 'p':
			strcpy(def_pal, optarg);
			break;
		}
	}

	if (optind == argc) {
		usage();
		exit(0);
	}
	/* Set 1024x768x256 video mode */

	if (vga_setmode(G1024x768x256) != 0) {
		printf("vga_setmode() failed\n");
		exit(4);
	}
	if (keyboard_init() != 0) {
		printf("Can't initialize keyboard RAW mode\n");
		vga_setmode(TEXT);
		exit(5);
	}
	keyboard_setdefaulteventhandler();

	/*
	 * Alloc font buffers
	 */
	small_font = malloc(8 * 8 * 256);
	normal_font = malloc(8 * 16 * 256);
	big_font = malloc(8 * 16 * 256);

	/*
	 * SVGALIB built-in 8x8 font, just in case...
	 */
	gl_setcontextvga(G1024x768x256);
	gl_expandfont(8, 8, 255, gl_font8x8, small_font);

	/*
	 * small font (8x8thin)
	 */
	if ((fp = fopen(FONTPATH "/hisquash-8", "rb")) != NULL) {
		fread(tmpbuf, 2048, 1, fp);
		fclose(fp);
		gl_expandfont(8, 8, 255, tmpbuf, small_font);
	}
	/*
	 * normal font (8x16alt)
	 */
	if ((fp = fopen(FONTPATH "/8x16alt", "rb")) != NULL) {
		have_normalfont = 1;
		fread(tmpbuf, 4096, 1, fp);
		fclose(fp);
		gl_expandfont(8, 16, 255, tmpbuf, normal_font);
	}
	/*
	 * big font (Roman 8x16)
	 */
	if ((fp = fopen(FONTPATH "/rl.fnt", "rb")) != NULL) {
		have_bigfont = 1;
		fread(tmpbuf, 4096, 1, fp);
		fclose(fp);
		gl_expandfont(8, 16, 255, tmpbuf, big_font);
	}
	/*
	 * read default palette
	 */
	if (*def_pal == '\0')
		read_palette("vn_gray.pal", &palette[0][0][0]);
	else
		read_palette(def_pal, &palette[0][0][0]);

	/*
	 * read palette #1
	 */
	read_palette("vn_color.pal", &palette[1][0][0]);

	/*
	 * set default palette
	 */
	for (il = 0; il < 256; il++) {
		gl_setpalettecolor(il, palette[0][il][0] / 4
				   ,palette[0][il][1] / 4
				   ,palette[0l][il][2] / 4);
	}

	if (have_bigfont)
		gl_setfont(8, 16, big_font);
	gl_write(0, 0, "NASA IMQ/IMG FILES VIEWER");
	gl_write(0, 20, version);
	gl_write(0, 45, "Press F1 for Help");
	gl_setfont(8, 8, small_font);

	for (end_all = 0; !end_all;) {
		strcpy(fname, argv[optind]);
		strcpy(temp, argv[optind]);

		filter[0] = '\0';
		target[0] = '\0';
		spacecraft[0] = '\0';
		date[0] = '\0';
		camera[0] = '\0';

		imqfile = ((strstr(fname, ".imq") != NULL) || (strstr(fname, ".IMQ") != NULL));
		imgfile = ((strstr(fname, ".img") != NULL) || (strstr(fname, ".IMG") != NULL));
		ibgfile = ((strstr(fname, ".ibg") != NULL) || (strstr(fname, ".IBG") != NULL));

		if (imqfile == 0 && imgfile == 0 && ibgfile == 0) {
			/*
			 * unknown file type, skip it
			 */
			optind++;
			if (optind == argc) {
				no_more_files(0, batch_mode);
				keyboard_close();
				vga_setmode(TEXT);
				free(small_font);
				free(normal_font);
				free(big_font);
				return 0;
			}
			continue;
		}
		/*
		 * IMG and IBG files do not have engineering table data and line
		 * suffix data, return to standard info display
		 */
		if ((imgfile) && (what_left > 2))
			what_left = 0;

		if ((ibgfile) && ((what_left == 3) || (what_left == 4)))
			what_left = 0;

		/*
		 * Check for detached label
		 */
		detached_label = 0;
		if (imgfile) {
			strcpy(lblname, fname);
			if ((cp = strrchr(lblname, '.')) != NULL)
				*cp = '\0';
			strcat(lblname, ".lbl");
			if ((fp = fopen(lblname, "rb")) != NULL) {
				detached_label = 1;
			}
		}
		if (!detached_label) {
			if ((fp = fopen(fname, "rb")) == NULL) {
				keyboard_close();
				vga_setmode(TEXT);
				free(small_font);
				printf("Can't open file \"%s\":%s\n", fname, strerror(errno));
				exit(6);
			}
		}
		/*
		 * read text info
		 */

		image_lines = 0;
		frecords = 0;
		nline = 0;
		lrecords = 0;
		lbytes = 0;
		hist_start = 0;
		image_start = 0;
		bytestoskip = 0;
		nbits = 8;
		npix = 0;
		record_bytes = 0;
		enhanced = 0;

		do {
			if (imqfile) {
				len = get_record(fp, tline, 0);
				tline[len] = '\0';
			} else {
				fgets(tline, 132, fp);
				len = strlen(tline);
			}

			cp = rip(tline);
			image_info[nline] = malloc(len + 2);
			strcpy(image_info[nline++], tline);
#ifdef DEBUG
			printf("(%2.2x) %s\n", len, cp);
#endif
			if (strncmp(cp, "RECORD_BYTES", 12) == 0)
				record_bytes = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "LABEL_RECORD_BYTES", 18) == 0)
				lbytes = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "LABEL_RECORDS", 13) == 0)
				lrecords = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "FILE_RECORDS", 12) == 0)
				frecords = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "IMAGE_LINES", 11) == 0)
				image_lines = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "LINES", 5) == 0)
				image_lines = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "LINE_PREFIX_BYTES", 17) == 0)
				bytestoskip = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "SAMPLE_BITS", 11) == 0)
				nbits = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "LINE_SAMPLES", 12) == 0)
				npix = atoi(strchr(cp, '=') + 2);
			if (strncmp(cp, "^IMAGE", 6) == 0) {
				if (detached_label)
					image_start = atoi(strchr(cp, ',') + 1);
				else
					image_start = atoi(strchr(cp, '=') + 2);
			}
			if (strncmp(cp, "^IMAGE_HISTOGRAM", 16) == 0) {
				if (detached_label)
					hist_start = atoi(strchr(cp, ',') + 1);
				else
					hist_start = atoi(strchr(cp, '=') + 2);
			}
			if (strncmp(cp, "TARGET_NAME", 11) == 0)
				strcpy(target, strchr(cp, '=') + 2);
			if (strncmp(cp, "SPACECRAFT_NAME", 15) == 0)
				strcpy(spacecraft, strchr(cp, '=') + 2);
			if (strncmp(cp, "IMAGE_TIME", 10) == 0)
				strcpy(date, strchr(cp, '=') + 2);
			if (strncmp(cp, "FILTER_NAME", 11) == 0)
				strcpy(filter, strchr(cp, '=') + 2);
			if (strncmp(cp, "INSTRUMENT_NAME", 11) == 0)
				strcpy(camera, strchr(cp, '=') + 2);
		}
		while (strcmp(cp, "END"));

#ifdef DEBUG
		printf("record_bytes = %d\n", record_bytes);
		printf("frecords = %d\n", frecords);
		printf("lrecords = %d\n", lrecords);
		printf("lbytes = %d\n", lbytes);
		printf("hist_start = %d\n", hist_start);
		printf("image_start = %d\n", image_start);
		printf("image_lines = %ld\n", image_lines);
		printf("npix = %d\n", npix);
		printf("nbits = %d\n", nbits);
		printf("spacecraft = %s\n", spacecraft);
		printf("target = %s\n", target);
		printf("date = %s\n", date);
#endif

		if (detached_label) {
			fclose(fp);
			if ((fp = fopen(fname, "rb")) == NULL) {
				free(small_font);
				vga_setmode(TEXT);
				printf("Can't open file \"%s\":%s\n", argv[1], strerror(errno));
				exit(6);
			}
		}
		/*
	 	 * get image histogram
 		 */
		if (imqfile) {
			len = get_record(fp, (char *) image_hist, 0);
#ifdef DEBUG
			printf("Len ih#1 = %d bytes\n", len);
#endif
			len = get_record(fp, (char *) image_hist + record_bytes, 0);
#ifdef DEBUG
			printf("Len ih#2 %d bytes\n", len);
#endif
		} else if (ibgfile) {
			fseek(fp, (long) (hist_start - 1) * (long) record_bytes, SEEK_SET);
			for (n = 0; n < 5; n++) {
				/*
				 * get the first 1000 bytes as 5 200-bytes records
				 */
				len = get_record(fp, (char *) image_hist + n * record_bytes, record_bytes);
#ifdef DEBUG
				printf("Len ih#%d = %d bytes\n", n, len);
#endif
			}
			/*
			 * get last 24 bytes
			 */
			len = get_record(fp, (char *) image_hist + 5 * record_bytes, 24);
		}
		if (imqfile) {
			/*
		 	 * get encoding histogram
	 		 */
			len = get_record(fp, (char *) encoding_hist, 0);
#ifdef DEBUG
			printf("Len eh#1 = %d bytes\n", len);
#endif
			len = get_record(fp, (char *) encoding_hist + record_bytes, 0);
#ifdef DEBUG
			printf("Len eh#2 = %d bytes\n", len);
#endif
			len = get_record(fp, (char *) encoding_hist + 2 * record_bytes, 0);
#ifdef DEBUG
			printf("Len eh#3 = %d bytes\n", len);
#endif
			/*
			 * get engineering summary
			 */
			len = get_record(fp, (char *) &engtab, 0);

#ifdef DEBUG
			printf("Len esum = %d bytes\n", len);
#endif
#ifdef DEBUG
			for (cp = (char *) &engtab, n = 0; n < 242; cp++, n++) {
				printf("%2.2X ", (unsigned char) *cp);
				if ((n + 1) % 16 == 0)
					printf("\n");
			}
#endif

			/*
			 * Pass the encoding_hist array to the decmpinit
			 * routine for intialization
	 		 */
			decmpinit(encoding_hist);
			nso = record_bytes;
		}
		/*
		 * The loop will read one compressed line at a time from
		 * the input file and call the decompress routine to restore
		 * the line
		 */
		gl_setclippingwindow(0, 745, 223, 767);
		gl_clearscreen(0);
		if (have_bigfont)
			gl_setfont(8, 16, big_font);
		gl_write(0, 750, "File:");
		if ((cp = strrchr(fname, '/')) == NULL)
			cp = fname;
		else
			cp++;

		gl_write(50, 750, cp);
		gl_setfont(8, 8, small_font);

		gl_setclippingwindow(223, 0, 1023, 767);

		if (clearscreen)
			gl_clearscreen(0);

		if (!imqfile) {
			if (image_start != 0) {
				if (ibgfile)
					fseek(fp, (long) (image_start - 1) * (long) record_bytes, SEEK_SET);
				else
					fseek(fp, (long) image_start * (long) record_bytes, SEEK_SET);
			} else if (lrecords != 0) {
				if (lbytes != 0)
					fseek(fp, (long) lrecords * (long) lbytes, SEEK_SET);
				else
					fseek(fp, (long) lrecords * (long) record_bytes, SEEK_SET);
			}
		}
		for (il = 0, y = 0; il < image_lines; il += 25) {
			for (n = 0; n < 25; n++) {
				if (imqfile) {
					nsi = get_record(fp, linei, 0);
					decompress(linei, &putbuf[npix * n], &nsi, &nso);
					memcpy((char *) &linsufx[il + n], &putbuf[npix * (n + 1)], 36);
				} else {
					get_record(fp, linei, record_bytes);
					convert(&linei[bytestoskip], &putbuf[npix * n], record_bytes, nbits);
				}
			}
			gl_putbox(224, y, npix, (disable_scaling) ? 25 : 24, putbuf);
			y += (disable_scaling) ? 25 : 24;
		}

		restore_left(what_left, nline, line, part, disable_clipping);

		for (end = 0; !end;) {
			for (now = then = clock(); (now - then) < 10; now = clock());
			if (autodisp) {
				for (now = then = clock(); ((ch = get_command()) == -1) && ((now - then) < (100l * delay)); now = clock());
				if (ch == -1)
					ch = 9;
			} else {
				while ((ch = get_command()) == -1);
			}
			switch (ch) {
			case 0:
				break;
			case 1:
				/*
				 * display help
				 */
				gl_setclippingwindow(0, 75, 223, 740);
				gl_clearscreen(0);
				what_left = 1;
				part = 0;
				help();
				break;
			case 11:
				/*
				 * display info
				 */
				disp_info(disable_clipping);
				what_left = 0;
				part = 0;
				break;
			case 2:
				/*
				 * display image data
				 */
				display_data(0, nline, disable_clipping);
				what_left = 2;
				part = 0;
				break;
			case 12:
				/*
				 * display image histogram
				 */
				if (!imqfile && !ibgfile)
					break;
				display_hist();
				what_left = 5;
				part = 0;
				break;
			case 3:
				/*
				 * display engineering table
				 */
				if (!imqfile)
					break;
				if (++part == 4)
					part = 1;
				disp_engtable(&engtab, part, disable_clipping);
				what_left = 3;
				break;
			case 14:
				/*
				 * display specified line data
				 */
				if (!imqfile)
					break;
				gl_setclippingwindow(0, 75, 223, 740);
				gl_clearscreen(0);
				dispsufx(linsufx, line, disable_clipping);
				gl_setclippingwindow(0, 700, 222, 740);
				graphgets("New line #: ", st, 3, 1);
				line = atoi(st) - 3;
				if (line < -2)
					line = -2;	/* will be line 0 */
				if (line >= image_lines)
					line = image_lines - 3;
				/* note fall-thru */
			case 4:
				/*
				 * display line suffix data
				 */
				if (!imqfile)
					break;
				gl_setclippingwindow(0, 75, 223, 740);
				gl_clearscreen(0);
				line += 2;
				if (line > image_lines - 1)
					line = 0;
				dispsufx(linsufx, line, disable_clipping);
				what_left = 4;
				part = 0;
				break;
			case 15:
				/*
				 * read palette 1
				 */
				gl_setclippingwindow(0, 700, 223, 740);
				gl_clearscreen(0);
				graphgets("File: ", st, 13, 0);

				if (*st == '\0')  {
					break;
				}

				if (read_palette(st, &palette[0][0][0]) == -1) {
					graphgets("File not found (hit any key)", st, 0, 0);
					restore_left(what_left, nline, line, part, disable_clipping);
					break;
				}
				restore_left(what_left, nline, line, part, disable_clipping);
				/* note fall_thru */
			case 5:/*
				* set palette 1
				*/
				for (il = 0; il < 256; il++) {
					gl_setpalettecolor(il, palette[0][il][0] / 4
						      ,palette[0][il][1] / 4
						    ,palette[0][il][2] / 4);
				}
				break;
			case 16:
				/*
				 * read palette 2
				 */
				graphgets("File: ", st, 13, 0);

				if (*st == '\0')  {
					break;
				}
				
				if (read_palette(st, &palette[1][0][0]) == -1) {
					graphgets("File not found (hit any key)", st, 0, 0);
					restore_left(what_left, nline, line, part, disable_clipping);
					break;
				}
				restore_left(what_left, nline, line, part, disable_clipping);
				/* note fall_thru */
			case 6:
				/*
				 * set palette 2
				 */
				for (il = 0; il < 256; il++) {
					gl_setpalettecolor(il, palette[1][il][0] / 4
						      ,palette[1][il][1] / 4
						    ,palette[1][il][2] / 4);
				}
				break;
			case 7:
				/*
				 * enhance gray scale
				 */
				if ((enhanced) || !(imqfile || ibgfile))
					break;

				gl_setclippingwindow(223, 0, 1023, 767);

				max = min = 0l;
				for (n = 0; n < 255; n++) {
					if (image_hist[n] > max) {
						max = image_hist[n];
					}
				}
				for (n = 0; n < 256; n++) {
					if (image_hist[n] < min) {
						min = image_hist[n];
					}
				}
				nmax = nmin = 0;
				for (n = 0; n < 256; n++) {
					if (image_hist[n] > (max / 100)) {
						nmax = n;
					}
				}
				for (n = 255; n >= 0; n--) {
					if (image_hist[n] > (max / 100)) {
						nmin = n;
					}
				}
#ifdef DEBUG
				printf("max=%d (%ld), min=%d (%ld)\n", nmax, image_hist[nmax], nmin, image_hist[nmin]);
#endif
				ymax = (image_lines > 768) ? 768 : image_lines;
				for (y = 0; y < ymax; y++) {
					cp = putbuf;
					gl_getbox(224, y, npix, 1, putbuf);
					for (x = 0; x < npix; x++) {
						oldpixel = *cp & 255;
						newpixel = (255 * (oldpixel - nmin)) / (nmax - nmin);
						if (newpixel > 255)
							newpixel = 255;
						if (newpixel < 0)
							newpixel = 0;
						*cp++ = newpixel;
					}
					gl_putbox(224, y, npix, 1, putbuf);

				}
				enhanced = 1;
				break;
			case 17:
				/*
				 * exp/log enhance contrast
				 */
				s = r = k = 0;
				do {
					graphgets("Enter k value: ", st, 6, 1);
					k = atof(st);
				} while (k <= 0);
				do {
					graphgets("Log or Exp (L/E): ", st, 3, 0);
				} while (toupper(*st) != 'L' && toupper(*st) != 'E');
				gl_setclippingwindow(223, 0, 1023, 767);
				ymax = (image_lines > 768) ? 768 : image_lines;
				for (y = 0; y < ymax; y++) {
					cp = putbuf;
					gl_getbox(224, y, npix, 1, putbuf);
					for (x = 0; x < npix; x++) {
						r = (float) (*cp & 255) / 255.0;
						switch (toupper(st[0])) {
						case 'L':
							s = (log(1 + (exp(k) - 1) * r) / k);
							break;
						case 'E':
							s = (pow((1 + k), r) - 1) / k;
							break;
						}
						newpixel = (int) (s * 255);
						if (newpixel > 255)
							newpixel = 255;
						if (newpixel < 0)
							newpixel = 0;
						*cp++ = newpixel;
					}
					gl_putbox(224, y, npix, 1, putbuf);

				}
				enhanced = 1;
				break;
			case 8:
				/*
				 * reload image
				 */
				end = 1;
				break;
			case 18:
				/*
				 * disable image scaling
				 */
				disable_scaling = ~disable_scaling;
				end = 1;
				break;
			case 9:
				/*
				 * show next file
				 */
				fclose(fp);
				if (imqfile)
					free_tree(&num);
				for (il = 0; il < nline; il++)
					free(image_info[il]);
				line = 0;
				optind++;
				if (optind == argc) {
					no_more_files(1, batch_mode);
					end = end_all = 1;
					break;
				}
				end = 1;
				break;
			case 19:
				/*
				 * automatic slideshow toggle
				 */
				autodisp = ~autodisp;
				if (autodisp)
					optind++;
				if (optind < argc) {
					end = 1;
					break;
				}
				no_more_files(1, batch_mode);
				/* note fall-thru */
			case 10:
				/*
				 * quit program
				 */
				fclose(fp);
				if (imqfile)
					free_tree(&num);
				for (il = 0; il < nline; il++)
					free(image_info[il]);
				end = 1;
				end_all = 1;
			}
		}
	}
	keyboard_close();
	vga_setmode(TEXT);
	free(small_font);
	free(normal_font);
	free(big_font);
	return 0;
}
