/********************************

 Various structures definitions

*********************************/

#ifndef _VN_H
#define _VN_H

/********************************
 Engineering table defines.
 For more details see ENGTAB.LBL
*********************************/

/* sub_structures defines */

typedef struct {
	unsigned short FIRST_ERT_DAY:9;
	unsigned short FIRST_ERT_YEAR:7;
	unsigned short FIRST_ERT_MINUTE;
	unsigned short FIRST_ERT_MILLISECOND;
} first_ert;

typedef struct {
	unsigned short LAST_ERT_DAY:9;
	unsigned short LAST_ERT_YEAR:7;
	unsigned short LAST_ERT_MINUTE;
	unsigned short LAST_ERT_MILLISECOND;
} last_ert;

typedef struct {
	unsigned short SCET_DAY:9;
	unsigned short SCET_YEAR:7;
	unsigned short SCET_MINUTE;
	unsigned short SCET_MILLISECOND;
} scet;


typedef struct {

	short SYNC_MSB;
	char SYNC_LSB;
	unsigned char SOURCE_ID;
	unsigned char DESTINATION_ID;
	unsigned char BLOCK_ID;
	unsigned short DDT_ID:6;
	unsigned short UDT_ID:7;
	unsigned short GDD_ID:3;
	unsigned short TIME_ID_MSB:8;
	unsigned short SPACECRAFT_NUMBER:7;
	unsigned short GCF_NOT_USED:1;
	unsigned short TIME_ID_LSB:16;
	unsigned short MSB_BLOCK_COUNT:2;
	unsigned short DAY:12;
	unsigned short DAY_NOT_USED:2;
	unsigned char LSB_BLOCK_COUNT;
	unsigned char MILL_COUNT;
	unsigned char SERIAL_NUMBER;
	unsigned char CONFIGURATION;
	unsigned short ESC:2;
	short ESC_NOT_USED:14;
} gcf_table;


typedef struct {
	unsigned short IRT_DAY:9;
	unsigned short IRT_YEAR:7;
	unsigned short IRT_MINUTE;
	unsigned short IRT_MILLISECOND;

} irt;

typedef struct {
	unsigned short FORMAT_SC_ID:1;
	unsigned short IMAGE_FORMAT_ID:5;
	unsigned short FORMAT_ID:2;
	unsigned short FORMAT_NOT_USED:8;
} format;

typedef struct {
	unsigned short SYNC_PARM_P:8;
	unsigned short SYNC_PARM_I:8;
} sync_parameter_1;


typedef struct {
	unsigned short SYNC_PARM_L:5;
	unsigned short SYNC_PARM_K:5;
	unsigned short SYNC_PARM_J:5;
	unsigned short SYNC_2_NOT_USED:1;
} sync_parameter_2;


typedef struct sync_parameter {
	unsigned short SYNC_PARM_R:5;
	unsigned short SYNC_PARM_N:5;
	unsigned short SYNC_PARM_M:5;
	unsigned short SYNC_3_NOT_USED:1;
} sync_parameter_3;


typedef	struct {
	unsigned short SHUTTERED_PICTURE_FLAG:6;
	unsigned short CAMERA_NUMBER:1;
	unsigned short SHUTTERED_NOT_USED:9;
} shuttered_picture_id;


typedef struct {
	unsigned READ_OUT_SEGMENT_NUMBER:6;
	unsigned short READ_OUT_LINE_NUMBER:10;
} scan_mode;

typedef struct {
	unsigned short FILTER_PARITY:1;
	unsigned short FILTER_ID:3;
	unsigned short EXPOSURE_ID:5;
	unsigned short WA_CALIBRATION_FLAG:1;
	unsigned short NA_CALIBRATION_FLAG:1;
	unsigned short CAMERA_NOT_USED:5;
} camera_mode;

typedef struct {
	unsigned char NA_ANALOG_SAMPLE;
	unsigned char WA_ANALOG_SAMPLE;
} analog_sample_table;


typedef struct {
	unsigned short SECONDARY_MEMORY_READOUT:1;
	unsigned short NA_LSB_TRUNC_FLAG:1;
	unsigned short WA_LSB_TRUNC_FLAG:1;
	unsigned short PIX_AVERAGE:5;
	unsigned short PIX_AVERAGE_STATUS:1;
	unsigned short PIX_NOT_USED:7;
} pix_stat;


/************************************** 
   And now, ladies and gentlemen...
   The Engineering Table structure !
***************************************/

typedef struct {
	unsigned char MTIS_FILE_NUMBER;
	unsigned char MTIS_RECORD_ID;
	unsigned short PHYSICAL_SEQUENCE_NUMBER;
	unsigned short LOGICAL_SEQUENCE_NUMBER;
	first_ert FIRST_ERT;
	last_ert LAST_ERT;
	short FIRST_FDS16_COUNT;
	short FIRST_FDS60_COUNT;
	short FIRST_FDS_LINE_COUNT;
	short LAST_FDS16_COUNT;
	short LAST_FDS60_COUNT;
	short LAST_FDS_LINE_COUNT;
	scet SCET;
	char MTIS_RECORDING_ID[32];
	gcf_table GCF_TABLE[2];
	irt IRT;
	long NOT_USED_A;
	format FORMAT;
	short MIN_BSNR;
	short MAX_BSNR;
	short MIN_SSNR;
	short MAX_SSNR;
	short MIN_AGC;
	short MAX_AGC;
	short SYNC_ERROR_COUNT;
	short FDS_ERROR_COUNT;
	sync_parameter_1 SYNC_PARAMETER_1;
	sync_parameter_2 SYNC_PARAMETER_2;
	sync_parameter_3 SYNC_PARAMETER_3;
	short LINES;
	short FULL_LINES;
	short PARTIAL_LINES;
	short UNREADABLE_RECORDS;
	short SEQUENCE_BREAKS;
	short SORT_PARAMETER[4];
	short IDR_FRAMES;
	short WBDL_FRAMES;
	short SDR_FRAMES;
	short MISSING_FRAMES;
	short NOT_USED_B;
	char IMAGE_ID[10];
	char TARGET_NAME[10];
	unsigned char INPUT_TYPE;
	unsigned char INPUT_SOURCE;
	shuttered_picture_id SHUTTERED_PICTURE_ID;
	scan_mode SCAN_MODE;
	camera_mode CAMERA_MODE;
	short PICTURE_COUNT;
	short PRESENT_VALUE_A;
	short WORD_INDICATOR_A;
	short WORD_POINTER_A;
	short PRESENT_VALUE_B;
	short WORD_POINTER_B;
	short PRESENT_VALUE_C;
	short WORD_INDICATOR_C;
	short PRESENT_VALUE_D;
	short WORD_INDICATOR_D;
	short WORD_POINTER_D;
	analog_sample_table ANALOG_SAMPLE_TABLE[5];
	pix_stat PIX_STAT;
	unsigned char ISS_ENG[10];
} engtable;


/**********************************
 The Line Suffix Table Structure
 For more details, see LINESUFX.LBL
***********************************/

typedef struct {
	short FDS_MOD16_NUMBER;
	short FDS_MOD60_NUMBER;
	short FDS_LINE_NUMBER;
	short MTIS_LINE_NUMBER;
	short MISSING_FRAMES;
	short RETAINED_FRAME_BITS[10];
	unsigned char INPUT_TYPE;
	unsigned char INPUT_SOURCE;
	short FIRST_SAMPLE_NUMBER;
	short LAST_SAMPLE_NUMBER;
} line_suffix;

#endif /* _VN_H */
